from django.shortcuts import reverse


class SimpleCrudAPITestCase(object):

    # Test settings
    DISABLE_DEFAULT_TESTS = False  # disable the default tests

    # Test Constants
    URL_NAMESPACE = None
    URL_NAME = None
    MODEL = None

    # Endpoint settings
    ACTIONS = [
        'list',
        'retrieve',
        'destroy',
        'create',
        'update',
        'partial_update',
    ]

    def check_configuration(self, method):

        # Make sure the class is setup correctly

        self.assertIsNotNone(
            self.URL_NAMESPACE,
            'Please specify the URL_NAMESPACE')
        self.assertIsNotNone(
            self.URL_NAME,
            'Please specify the URL_NAME')
        self.assertIsNotNone(
            self.MODEL,
            'Please specify the MODEL')
        # Check the options for ACTIONS
        for a in self.ACTIONS:
            self.assertIn(
                a,
                [
                    'list',
                    'retrieve',
                    'destroy',
                    'create',
                    'update',
                    'partial_update',
                ],
                '{} is not a valid action name'.format(a)
            )

        # Lastly check the method is allowed
        return method in self.ACTIONS

    def get_reversed_url(self, action, *args):
        if action in [
            'retrieve',
            'destroy',
            'update',
            'partial_update'
        ]:
            suffix = 'detail'
        else:
            suffix = 'list'

        url = '{}:{}-{}'.format(self.URL_NAMESPACE, self.URL_NAME, suffix)
        reversed_url = reverse(url, args=args)
        return reversed_url

    def create_test_object(self):
        """
        Creates a test object for the model viewset
        """
        return self.MODEL.objects.create(**self.get_model_payload())

    def get_model_payload(self):  # pragma: no cover
        raise NotImplementedError("This needs to be abstracted")

    def get_request_payload(self):  # pragma: no cover
        raise NotImplementedError("This needs to be abstracted")

    def get_request_headers(self):
        return {}

    # Test retrieve list
    def test_can_read_list(self):

        # Check if default tests are disabled
        if self.DISABLE_DEFAULT_TESTS:  # pragma: no cover
            return

        # - First check that the class is setup and assign the correct status
        status = 200 if self.check_configuration('list') else 405

        # get the URL
        url = self.get_reversed_url('list')

        # Make the request
        response = self.client.get(url, headers=self.get_request_headers())

        # Assert statuscode
        self.assertEqual(response.status_code, status, response.json())

    # Test retrieve detail
    def test_can_read_retrieve(self):

        # Check if default tests are disabled
        if self.DISABLE_DEFAULT_TESTS:  # pragma: no cover
            return

        # - First check that the class is setup and assign the correct status
        status = 200 if self.check_configuration('retrieve') else 405

        # Create the test object
        test_obj = self.create_test_object()

        # get the URL
        url = self.get_reversed_url('retrieve', test_obj.id)

        # Make the request
        response = self.client.get(url, headers=self.get_request_headers())

        # Assert statuscode
        self.assertEqual(response.status_code, status, response.json())

    # Test write create
    def test_can_write_create(self):

        # Check if default tests are disabled
        if self.DISABLE_DEFAULT_TESTS:  # pragma: no cover
            return

        # - First check that the class is setup and assign the correct status
        status = 201 if self.check_configuration('create') else 405

        # get the URL
        url = self.get_reversed_url('create')

        payload = self.get_request_payload()

        # Make the request
        response = self.client.post(url, headers=self.get_request_headers())

        # Assert statuscode
        self.assertEqual(response.status_code, status, response.json())

    # Test write update
    def test_can_write_update(self):

        # Check if default tests are disabled
        if self.DISABLE_DEFAULT_TESTS:  # pragma: no cover
            return

        # - First check that the class is setup and assign the correct status
        status = 200 if self.check_configuration('update') else 405

        # get the URL
        test_obj = self.create_test_object()
        url = self.get_reversed_url('update', test_obj.id)
        payload = self.get_request_payload()

        # Make the request
        response = self.client.put(url, headers=self.get_request_headers())

        # Assert statuscode
        self.assertEqual(response.status_code, status, response.json())

    # Test write partial_update
    def test_can_write_partial_update(self):

        # Check if default tests are disabled
        if self.DISABLE_DEFAULT_TESTS:  # pragma: no cover
            return

        # - First check that the class is setup and assign the correct status
        status = 200 if self.check_configuration('partial_update') else 405

        # get the URL
        test_obj = self.create_test_object()
        url = self.get_reversed_url('partial_update', test_obj.id)
        payload = self.get_request_payload()

        # Make the request
        response = self.client.patch(url, headers=self.get_request_headers())

        # Assert statuscode
        self.assertEqual(response.status_code, status, response.json())

    # Test write destroy
    def test_can_write_destroy(self):

        # Check if default tests are disabled
        if self.DISABLE_DEFAULT_TESTS:  # pragma: no cover
            return

        # - First check that the class is setup and assign the correct status
        status = 204 if self.check_configuration('destroy') else 405

        # get the URL
        test_obj = self.create_test_object()
        url = self.get_reversed_url('destroy', test_obj.id)
        payload = self.get_request_payload()

        # Make the request
        response = self.client.delete(url, headers=self.get_request_headers())

        # Assert statuscode
        self.assertEqual(response.status_code, status)



from django.db import models
from django.utils import timezone
import uuid
from .mixins import ModelDiffMixin


class BaseQuerySet(models.QuerySet):
    """
    A base queryset that has some extended functionality from django
    """

    def delete(self, force=False):
        """
        Fakes a queryset .delete() and instead adds the time an item has
        been deleted

        Keyword Arguments:
            force {bool} -- Flag on whether to use a hard delete and delete
            the item from the database, or to do a soft delete instead
        """
        # Delete all items anyway
        if force:
            super().delete()

        # Perform a soft delete instead
        else:
            return self.update(deleted=timezone.now())

    def deleted(self):
        """
        Gets all objects that are deleted
        """
        return self.filter(deleted__isnull=False)

    def active(self):
        '''
        Return all stock that is not deleted
        '''
        return self.filter(deleted__isnull=True)

    def undelete(self):
        return self.update(deleted=None)


class BaseManager(models.Manager):

    """
    A base queryset that has some extended functionality from django
    """

    def get_queryset(self):
        return BaseQuerySet(self.model, using=self._db)

    def deleted(self):
        """
        Returns only deleted items        
        """
        return self.get_queryset().deleted()

    def all(self):
        """
        When using MODEL.objects.all() -> Instead of getting all objects,
        rather only get objects that are not deleted.
        """
        return self.get_queryset().active()


class CoreModel(ModelDiffMixin, models.Model):
    """
    A base model that has repeatable behaviour that
    all other models should inherit
    """

    # -- Set the object manager
    objects = BaseManager()

    # -- Default fields
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)
    deleted = models.DateTimeField(null=True, blank=True, editable=False)
    uuid = models.UUIDField(default=uuid.uuid4)

    class Meta:
        abstract = True

    def delete(self, force=False):
        """Performs a delete on the object

        Keyword Arguments:
            force {bool} -- Flag on whether to use a hard delete and delete
            the item from the database, or to do a soft delete instead
        """

        if force:
            # Perform a hard delete
            super().delete()

        else:
            # Perform a soft delete
            self.deleted = timezone.now()
            self.save()

    def undelete(self):
        """Remove the `deleted` value and undo the soft delete"""

        self.deleted = None
        self.save()

    def save(self, *a, **k):
        # -- Run pre save
        self.pre_save()

        # -- Run actual save
        super().save(*a, **k)

        # -- update our variables
        self._is_created = self.existing_instance is None
        self._is_updated = True

        # run post save
        self.post_save()

    def pre_save(self):
        """
        Pre Save Trigger, you can use this to change instance variables
        or even add/alter or run pre-save signals
        """
        pass

    def post_save(self):
        """
        Post Save Trigger, called after .save() and can be used for any 
        post processing. Similar to signals.
        """
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_updated = False
        self.is_new_instance = self.pk is None

    @property
    def is_deleted(self):
        return bool(self.deleted)

    @property
    def is_created(self):
        return bool(self.existing_instance)

    @property
    def is_updated(self):
        return self._is_updated

    @property
    def existing_instance(self):
        def _call():
            if self.pk:
                return self.__class__.objects.get(pk=self.pk)
        return _call()


# class ImageGallery(models.Model):
#     image = models.ImageField(
#         upload_to='', blank=True,
#         null=True
#     )
#     caption = models.CharField(max_length=100, blank=True, null=True)

#     def __str__(self):
#         return self.image.url



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['django_handyman', 'django_handyman.tests']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'django-handyman',
    'version': '1.0.2rc1',
    'description': '',
    'long_description': None,
    'author': 'Giannis Katsini',
    'author_email': 'giannis.katsini@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
